print("import: 'moto'")
import moto

print("import: 'moto.acm'")
import moto.acm

print("import: 'moto.apigateway'")
import moto.apigateway

print("import: 'moto.autoscaling'")
import moto.autoscaling

print("import: 'moto.awslambda'")
import moto.awslambda

print("import: 'moto.batch'")
import moto.batch

print("import: 'moto.cloudformation'")
import moto.cloudformation

print("import: 'moto.cloudwatch'")
import moto.cloudwatch

print("import: 'moto.cognitoidentity'")
import moto.cognitoidentity

print("import: 'moto.cognitoidp'")
import moto.cognitoidp

print("import: 'moto.config'")
import moto.config

print("import: 'moto.core'")
import moto.core

print("import: 'moto.datapipeline'")
import moto.datapipeline

print("import: 'moto.dynamodb'")
import moto.dynamodb

print("import: 'moto.dynamodb2'")
import moto.dynamodb2

print("import: 'moto.dynamodbstreams'")
import moto.dynamodbstreams

print("import: 'moto.ec2'")
import moto.ec2

print("import: 'moto.ec2.responses'")
import moto.ec2.responses

print("import: 'moto.ecr'")
import moto.ecr

print("import: 'moto.ecs'")
import moto.ecs

print("import: 'moto.elb'")
import moto.elb

print("import: 'moto.elbv2'")
import moto.elbv2

print("import: 'moto.emr'")
import moto.emr

print("import: 'moto.events'")
import moto.events

print("import: 'moto.glacier'")
import moto.glacier

print("import: 'moto.glue'")
import moto.glue

print("import: 'moto.iam'")
import moto.iam

print("import: 'moto.instance_metadata'")
import moto.instance_metadata

print("import: 'moto.iot'")
import moto.iot

print("import: 'moto.iotdata'")
import moto.iotdata

print("import: 'moto.kinesis'")
import moto.kinesis

print("import: 'moto.kms'")
import moto.kms

print("import: 'moto.logs'")
import moto.logs

print("import: 'moto.opsworks'")
import moto.opsworks

print("import: 'moto.organizations'")
import moto.organizations

print("import: 'moto.packages'")
import moto.packages

print("import: 'moto.packages.httpretty'")
import moto.packages.httpretty

print("import: 'moto.polly'")
import moto.polly

print("import: 'moto.rds'")
import moto.rds

print("import: 'moto.rds2'")
import moto.rds2

print("import: 'moto.redshift'")
import moto.redshift

print("import: 'moto.resourcegroupstaggingapi'")
import moto.resourcegroupstaggingapi

print("import: 'moto.route53'")
import moto.route53

print("import: 'moto.s3'")
import moto.s3

print("import: 'moto.s3bucket_path'")
import moto.s3bucket_path

print("import: 'moto.secretsmanager'")
import moto.secretsmanager

print("import: 'moto.ses'")
import moto.ses

print("import: 'moto.sns'")
import moto.sns

print("import: 'moto.sqs'")
import moto.sqs

print("import: 'moto.ssm'")
import moto.ssm

print("import: 'moto.sts'")
import moto.sts

print("import: 'moto.swf'")
import moto.swf

print("import: 'moto.swf.models'")
import moto.swf.models

print("import: 'moto.xray'")
import moto.xray

